;;; This source file is part of the Meta-CVS program, 
;;; which is distributed under the GNU license.
;;; Copyright 2002 Kaz Kylheku

(require "system")
(require "mapping")
(require "options")
(provide "filt")

(defun make-filt-hash (mapping)
  (let ((h (make-hash-table :test #'equal)))
    (dolist (entry mapping h)
      (multiple-value-bind (suffix nosuffix)
			   (suffix (mapping-entry-id entry))
	(declare (ignore suffix))
	(setf (gethash nosuffix h) entry)))))

(defun filt-select-map (filt-options)
  (find-bind (:test #'string= :key #'first :take #'second)
	     ((revision "r") (date "D") (extra-r "r") (extra-d "D")) 
	     filt-options
    (cond
      ((or extra-r extra-d)
	 (error "only one date or revision may be specified."))
      ((or revision date)
	 (with-input-from-program (stream `("cvs" "-Q" "up" "-p" 
					    ,@(format-opt filt-options)
					    ,*mcvs-map*))
	   (mapping-read stream)))
      (t (mapping-read *mcvs-map-local*)))))

(defun mcvs-filt (filt-options)
  (in-sandbox-root-dir
    (let* ((filemap (filt-select-map filt-options))
	   (filehash (make-filt-hash filemap)))
      (loop
	(let ((line (read-line t nil)))	
	  (when (null line) 
	    (return (values)))
	  (loop
	    (let ((f-start (search "F-" line :test #'char=))
		  (mcvs-dir (search "MCVS/F-" line :test #'char=)))
	      (flet ((is-hex-digit (x) 
		       (or (digit-char-p x)
			   (find x "ABCDEF"))))
		(cond
		  ((and mcvs-dir (or (and f-start (< mcvs-dir f-start))
				     (not f-start)))
		    (write-string (substring line 0 (+ mcvs-dir 7)))
		    (setf line (substring line (+ mcvs-dir 7))))
		  (f-start
		    (write-string (substring line 0 f-start))
		    (setf line (substring line (+ f-start 2)))
		    (when (< (length line) 32)
		      (write-string "F-")
		      (write-line line)
		      (return))
		    (cond 
		      ((notevery #'is-hex-digit (substring line 0 32))
			 (write-string "F-")
			 (setf line (substring line 2)))
		      (t (let* ((f-digits (substring line 0 32))
				(entry (gethash (format nil "F-~a" f-digits)
						filehash))
				(suffix (and entry 
					     (suffix (mapping-entry-id entry)))))
			   (setf line (substring line 32))
			   (cond
			     ((null entry)
				(write-string "F-")
				(write-string f-digits))
			     ((and suffix 
				   (or (< (length line) (1+ (length suffix)))
				       (not (path-equal (substring line 1 
								   (1+ (length suffix)))
							suffix))))
				(write-string "F-")
				(write-string f-digits))
			     (t (write-string (mapping-entry-path entry))
				(when suffix
				  (setf line 
					(substring line 
						   (1+ (length suffix)))))))))))
		  (t (write-line line)
		     (return)))))))))))

(defun mcvs-filt-wrapper (cvs-options cvs-command-options mcvs-args)
  (declare (ignore cvs-options))
  (when mcvs-args
    (error "no arguments permitted."))
  (mcvs-filt cvs-command-options))
